;; Generated package description from tramp-hlo.el  -*- no-byte-compile: t -*-
(define-package "tramp-hlo" "0.0.2" "High level operations as Tramp handlers" '((tramp "2.8.0.5")) :commit "a204c14a2eed4c92a455f12acae2ec7eeac6d8fd" :authors '(("Joe Sadusk" . "joe@sadusk.com")) :maintainer '("Joe Sadusk" . "joe@sadusk.com") :url "https://github.com/jsadusk/tramp-hlo")
